/*
 *  qoob mp3 player 
 *
 *  (c) qoob 
 *
 *  licensed as GPL
 *
 */


/* 
   Limitations: 
    - only suport 44.1kHz mp3 files
    - mono support is broken (vis.)
    - you can't skip/seek in the mp3
    - no shuffle support
*/

#include <string.h>
#include <mad/mad.h>
#include "cache.h"
#include <processor.h>

#define log __ieee754_log
#define sqrtf __ieee754_sqrtf
#include <math.h>
#include "dvd/dvd.h"

#define MEM_FB 0xC0f00000
#define MEM_WORK 0x81100000
#define MEM_FONT 0x81120000

unsigned long *fb = (void*)MEM_FB;

/* exi */
#define EXI_SR                *(volatile unsigned long*) 0xCC006800
#define EXI_DMA_MEM           *(volatile unsigned long*) 0xCC006804
#define EXI_DMA_LEN           *(volatile unsigned long*) 0xCC006808
#define EXI_CR                *(volatile unsigned long*) 0xCC00680C
#define EXI_DATA              *(volatile unsigned long*) 0xCC006810
#define EXI_WAIT_EOT          while((EXI_CR)&1);    

void exi_select(void)
{
	EXI_SR |= 0x150;
}

void exi_deselect(void)
{
	EXI_SR &= ~0x100;
}

void exi_write_word(unsigned long word)
{
	EXI_DATA = word;
	EXI_CR = 0x35;
	EXI_WAIT_EOT;
}

void exi_write(unsigned char *data, int len)
{
	while (len)
	{
		int l = len;
		if (l > 4)
			l = 4;
		EXI_DATA = *(unsigned long*)data;
		EXI_CR = 0x05 | ((l-1)<<4);
		EXI_WAIT_EOT;
		
		data += l;
		len -= l;
	}
}

void exi_read(unsigned char *dst, int len)
{
	while (len)
	{
		int l = len;
		if (l > 4)
			l = 4;
		EXI_DATA = 0;
		EXI_CR = 0x31;
		EXI_WAIT_EOT;
		*(unsigned long*)dst = EXI_DATA;
		dst += 4;
		len -= l;
	}
}

/* vi */

static const unsigned long video_ntsc[32] = {
0x0F060001, 0x476901AD, 0x02EA5140, 0x00030018,
0x00020019, 0x410C410C, 0x40ED40ED, 0x00000000,
0x00000000, 0x00000000, 0x00000000, 0x00000000,
0x110701AE, 0x10010001, 0x00010001, 0x00010001,
0x00000000, 0x00000000, 0x28500100, 0x1AE771F0,
0x0DB4A574, 0x00C1188E, 0xC4C0CBE2, 0xFCECDECF,
0x13130F08, 0x00080C0F, 0x00FF0000, 0x00000000,
0x02800000, 0x000000FF, 0x00FF00FF, 0x00FF00FF};

static const unsigned long video_pal[32] = {
0x11F50101, 0x4B6A01B0, 0x02F85640, 0x00010023,
0x00000024, 0x4D2B4D6D, 0x4D8A4D4C, 0x00000000,
0x00000000, 0x00000000, 0x00000000, 0x013C0144,
0x113901B1, 0x10010001, 0x00010001, 0x00010001,
0x00000000, 0x00000000, 0x28500100, 0x1AE771F0,
0x0DB4A574, 0x00C1188E, 0xC4C0CBE2, 0xFCECDECF,
0x13130F08, 0x00080C0F, 0x00FF0000, 0x00000000,
0x02800000, 0x000000FF, 0x00FF00FF, 0x00FF00FF};


void vi_init(unsigned long *mode, unsigned long fbaddr)
{
	int i;
	*(volatile unsigned long*) 0xCC002000 = 0x00000003;
		
	mode[7] = fbaddr;
	mode[9] = fbaddr + 1280;
	for (i=0; i<32; ++i)
		((unsigned long*)0xCC002000)[i] = mode[i];
}

/* Yay0 decompression */
void yay0_decode(void *s, void *d)
{
	int i, j, k, p, q, cnt;

	i = *(unsigned long *)(s + 4);	  // size of decoded data
	j = *(unsigned long *)(s + 8);	  // link table
	k = *(unsigned long *)(s + 12);	 // byte chunks and count modifiers

	q = 0;					// current offset in dest buffer
	cnt = 0;				// mask bit counter
	p = 16;					// current offset in mask table

	unsigned long r22 = 0, r5;
	
	do
	{
		// if all bits are done, get next mask
		if(cnt == 0)
		{
			// read word from mask data block
			r22 = *(unsigned long *)(s + p);
			p += 4;
			cnt = 32;   // bit counter
		}
		// if next bit is set, chunk is non-linked
		if(r22 & 0x80000000)
		{
			// get next byte
			*(unsigned char *)(d + q) = *(unsigned char *)(s + k);
			k++, q++;
		}
		// do copy, otherwise
		else
		{
			// read 16-bit from link table
			int r26 = *(unsigned short *)(s + j);
			j += 2;
			// 'offset'
			int r25 = q - (r26 & 0xfff);
			// 'count'
			int r30 = r26 >> 12;
			if(r30 == 0)
			{
				// get 'count' modifier
				r5 = *(unsigned char *)(s + k);
				k++;
				r30 = r5 + 18;
			}
			else r30 += 2;
			// do block copy
			unsigned char *pt = ((unsigned char*)d) + r25;
			int i;
			for(i=0; i<r30; i++)
			{
				*(unsigned char *)(d + q) = *(unsigned char *)(pt - 1);
				q++, pt++;
			}
		}
		// next bit in mask
		r22 <<= 1;
		cnt--;

	} while(q < i);
}

void ipl_read(unsigned char *dst, int address, int len)
{
	while (len)
	{
		int l = len;
		if (l > 0x100)
			l = 0x100;
		
		exi_select();
		exi_write_word(address << 6);
		exi_read(dst, l);
		exi_deselect();
		dst += l;
		address += l;
		len -= l;
	}
}

void untile(unsigned char *dst, unsigned char *src, int xres, int yres)
{
		// 8x8 tiles
	int x, y;
	int t=0;
	for (y = 0; y < yres; y += 8)
		for (x = 0; x < xres; x += 8)
		{
			t = !t;
			int iy, ix;
			for (iy = 0; iy < 8; ++iy, src+=2)
			{
				unsigned char *d = dst + (y + iy) * xres + x;
				for (ix = 0; ix < 2; ++ix)
				{
					int v = src[ix];
					*d++ = ((v>>6)&3);
					*d++ = ((v>>4)&3);
					*d++ = ((v>>2)&3);
					*d++ = ((v)&3);
				}
			}
		}
}

int font_offset[256], font_size[256], font_height;

void init_font(void)
{
	ipl_read((unsigned char*)MEM_FONT, 0x1FCF00, 0x3000);
	
	yay0_decode((void*)MEM_FONT, (void*)MEM_WORK);
	
	struct font_hdr
	{
		unsigned short font_type, first_char, last_char, subst_char, ascent_units, descent_units, widest_char_width,
			leading_space, cell_width, cell_height;
		unsigned long texture_size;
		unsigned short texture_format, texture_columns, texture_rows, texture_width, texture_height, offset_charwidth;
		unsigned long offset_tile, size_tile;
	} *fnt = (void*)MEM_WORK;
	
	untile((void*)MEM_FONT, (void*)(MEM_WORK + fnt->offset_tile), fnt->texture_width, fnt->texture_height);
	
	int i;
	for (i=0; i<256; ++i)
	{
		int c = i;
		if ((c < fnt->first_char) || (c > fnt->last_char))
			c = fnt->subst_char;
		else
			c -= fnt->first_char;

		font_size[i] = ((unsigned char*)fnt)[fnt->offset_charwidth + c];

		int r = c / fnt->texture_columns;
		c %= fnt->texture_columns;
		
		font_offset[i] = (r * fnt->cell_height) * fnt->texture_width + (c * fnt->cell_width);
	}
	
	font_height = fnt->cell_height;
}

long blit_lookup[4]    ={0x258e2573, 0x6d896d77, 0xb584b57b, 0xff80ff80};
long blit_lookup_inv[4]={0xff80ff80, 0xb584b57b, 0x6d896d77, 0x258e2573};

void blit_char(int x, int y, unsigned char c, unsigned long *lookup)
{
	unsigned char *fnt = ((unsigned char*)MEM_FONT) + font_offset[c];
	int ay, ax;
	for (ay=0; ay<font_height; ++ay)
	{
		for (ax=0; ax<font_size[c]; ax++)
		{
			int v0 = fnt[ax];
			unsigned long o = fb[(ay+y)*320+(ax+x)/2];
			
			if ((ax+x) & 1)
			{
				o &= ~0x00FFFFFF;
				o |= lookup[v0] & 0x00FFFFFF;
			} else
			{
				o &= ~0xFF000000;
				o |= lookup[v0] & 0xFF000000;
			}
			fb[(ay+y)*320+(ax+x)/2] = o;
		}
		fnt += 512;
	}
}

void write_font(int x, int y, const unsigned char *string)
{
	while (*string)
	{
		blit_char(x, y, *string, blit_lookup);
		x += font_size[*string];
		
		string++;
	}
}

void writex(int x, int y, int sx, int sy, const unsigned char *string, unsigned long *lookup)
{
	int ox = x;
	while ((*string) && ((x+16) < (ox + sx)))
	{
		blit_char(x, y, *string, lookup);
		x += font_size[*string];
		
		string++;
	}
	
	int ay;
	for (ay=0; ay<sy; ay++)
	{
		int ax;
		for (ax=x; ax<(ox + sx); ax += 2)
			fb[(ay+y)*320+ax/2] = lookup[0];
	}
}

void dvd_do_hack();

void set_status(char *message)
{
	writex(120, 80, 640-120, font_height, message, blit_lookup);
}

void set_video(int w)
{
	vi_init(w ? video_pal : video_ntsc, MEM_FB);
}

void video_set_framebuffer(int w, unsigned long *fb)
{
	((unsigned long*)0xCC002000)[w ? 9 : 7] = fb;
}

extern void _start(), _end();

struct pvd_s
{
	char id[8];
	char system_id[32];
	char volume_id[32];
	char zero[8];
	unsigned long total_sector_le, total_sect_be;
	char zero2[32];
	unsigned long volume_set_size, volume_seq_nr;
	unsigned short sector_size_le, sector_size_be;
	unsigned long path_table_len_le, path_table_len_be;
	unsigned long path_table_le, path_table_2nd_le;
	unsigned long path_table_be, path_table_2nd_be;
	unsigned char root_direntry[34];
	char volume_set_id[128], publisher_id[128], data_preparer_id[128], application_id[128];
	char copyright_file_id[37], abstract_file_id[37], bibliographical_file_id[37];
		// some additional dates, but we don't care for them :)
}  __attribute__((packed));


unsigned char sector_buffer[2048] __attribute__((aligned(32)));

int read_sector(void *buffer, int sector)
{
	return dvd_read(buffer, 2048, sector * 2048);
}

void fatal_read_error()
{
	set_status("FATAL READ ERROR! sorry.");
	while(1);
}

int is_unicode;

struct
{
	char name[128];
	int flags;
	int sector, size;
} file[1024];

int files;

int last_current_dir = -1;

int read_direntry(unsigned char *direntry)
{
	int nrb = *direntry++;
	++direntry;
	
	int sector;

	direntry += 4;
	sector  = (*direntry++) << 24;
	sector |= (*direntry++) << 16;
	sector |= (*direntry++) << 8;
	sector |= (*direntry++);	
	
	int size;
	
	direntry +=4;
	
	size  = (*direntry++) << 24;
	size |= (*direntry++) << 16;
	size |= (*direntry++) << 8;
	size |= (*direntry++);
	
	direntry += 7; // skip date
	
	int flags = *direntry++;
	++direntry; ++direntry; direntry+=4;
	
	int nl = *direntry++;
	
	char *name = file[files].name;

	file[files].sector = sector;
	file[files].size = size;
	file[files].flags = flags;
	
	if ((nl == 1) && (direntry[0] == 1)) // ".."
	{
		file[files].name[0]=0;
		if (last_current_dir != sector)
			files++;
	} else if ((nl == 1) && (direntry[0] == 0))
	{
		last_current_dir = sector;
	} else
	{
		if (is_unicode)
		{
			int i;
			for (i=0; i<(nl/2); ++i)
				name[i] = direntry[i*2+1];
			name[i]=0;
			nl = i;
		} else
		{
			memcpy(name, direntry, nl);
			name[nl]=0;
		}
		
		if (!(flags & 2))
		{
			if (name[nl - 2] == ';')
				name[nl-2] = 0;
			
			int i = nl;
			while (i >= 0)
				if (name[i] == '.')
					break;
				else
					--i;
			
			++i;
			
			int ext = i;
			
			char *cmp = "MP\x13";
			while (name[i])
				if ((name[i++] &~0x20) != *cmp++)
					break;
			
			if (*cmp || name[i])
				return nrb;
		} else
		{
			name[nl++] = '/';
			name[nl] = 0;
		}

		files++;
	}
	
	return nrb;
}

void read_directory(int sector, int len)
{
	read_sector(sector_buffer, sector);
	
	int ptr = 0;
	files=0;
	
	while (len > 0)
	{
		ptr += read_direntry(sector_buffer + ptr);
		if (!sector_buffer[ptr])
		{
			len -= 2048;
			read_sector(sector_buffer, ++sector);
			ptr = 0;
		}
	}
}

#define PAGESIZE   15

void show_files(int offset, int selection)
{
	int x = 100;
	int y = 100;
	
	int i;
	for (i=offset; (i < (offset + PAGESIZE)) && (i < files); ++i, y+=font_height)
	{
		writex(x, y, 400, font_height, file[i].name[0] ? file[i].name : "..", (selection  == i) ? blit_lookup_inv :  blit_lookup);
	}
	
	while (i < (offset + PAGESIZE))
	{
		writex(x, y, 400, font_height, "", blit_lookup);
		y += font_height;
		++i;
	}
}

int audio_len;

void audio_start(void *addr, int len)
{
	*(unsigned long*)0xCC006c04 = 0xFFFF; // max. volume
	
	*(unsigned short*)0xCC005030 = (((int)addr)&0x1FF0000) >> 16;
	*(unsigned short*)0xCC005032 = ((int)addr) & 0xFFFF;
	*(unsigned short*)0xCC005036 = (len/32) | 0x8000;
	
	*(unsigned long*)0xCC006c00 = 1; // start playing
	
	audio_len = len;
}

void audio_stop(void)
{
	*(unsigned long*)0xCC006c00 = 0;
	*(unsigned long*)0xCC006c04 = 0;
	*(unsigned short*)0xCC005036 = 0;
	
	*(unsigned long*)0xCC006c00 = 1; // start playing
}

int audio_get_pos()
{
	int rem = (*(volatile unsigned short*)0xCC00503a)*32;
	if (!rem)
		rem = audio_len;
	return audio_len - rem;
}

int read_safe(void *dst, int offset, int len)
{
	int ol = len;
	while (len)
	{
		int sector = offset / 2048;
		read_sector(sector_buffer, sector);
		int off = offset & 2047;
		
		int rl = 2048 - off;
		if (rl > len)
			rl = len;
		memcpy(dst, sector_buffer + off, rl);
		
		offset += rl;
		len -= rl;
		dst += rl;
	}
	
	return ol;
}

#define INPUT_BUFFER_SIZE 5*8192
unsigned char inputbuffer[INPUT_BUFFER_SIZE];

static inline signed short MadFixedToSshort(mad_fixed_t Fixed)
{
	Fixed=Fixed>>(MAD_F_FRACBITS-14);
	return((signed short)Fixed);
}

#define AUDIO_SIZE	 65536
#define MEM_AUDIO           (void*)0x80e00000

#define FRAC    8192
#define FRACSHIFT 13

/**************************************************************************************************/

int current_offset;
struct mad_stream stream;
struct mad_frame frame;
struct mad_synth synth;
mad_timer_t timer;
int status = 0, framecount = 0, remaining_size;

unsigned char *wptr = MEM_AUDIO, *base = MEM_AUDIO, *end = MEM_AUDIO + AUDIO_SIZE;

int fracptr = 0;
signed short sample[2][2];

		/* this refills the audio buffer, so it works by polling. */
		/* we assume that polling once per frame is enough. please don't */
		/* consume more time than one frame or gfx won't be smooth. */
static void refill_audio_buffer(void)
{
	while (!status)
	{
		int _wptr = (int)wptr - (int)MEM_AUDIO;
	
		int avail = _wptr - audio_get_pos();
		if (avail < 0)
			avail += AUDIO_SIZE;
		
		if (avail > 48000)
			break;
		
		if ((!stream.buffer) || (stream.error == MAD_ERROR_BUFLEN))
		{
			unsigned char *read_start;
			int read_size, remaining;
			if (stream.next_frame)
			{
				remaining = stream.bufend - stream.next_frame;
				memmove(inputbuffer, stream.next_frame, remaining);
				read_start = inputbuffer + remaining;
				read_size = INPUT_BUFFER_SIZE - remaining;
			} else
			{
				read_size = INPUT_BUFFER_SIZE;
				read_start = inputbuffer;
				remaining = 0;
			}
			
			if (read_size > remaining_size)
				read_size = remaining_size;
			
			read_size = read_safe(read_start, current_offset, read_size);
			current_offset += read_size;
			remaining_size -= read_size;
			
			mad_stream_buffer(&stream, inputbuffer, read_size + remaining);
			stream.error = 0;
		}
		
		if (mad_frame_decode(&frame, &stream))
		{
			if (MAD_RECOVERABLE(stream.error))
			{
//				if (stream.error != MAD_ERROR_LOSTSYNC) // || stream.this_frame != guardptr)
//					printf("recoverable frame level error (%s)\n", mad_stream_errorstr(&stream));
				continue;
			} else if (stream.error == MAD_ERROR_BUFLEN)
				continue;
			else {
//				printf("unrecoverable frame level error (%s)\n", mad_stream_errorstr(&stream));
				status = 1; /* unrec. error */
			}
		}
		
		if (!framecount)
		{
			// print info
		}
		framecount++;

		mad_timer_add(&timer, frame.header.duration);
		mad_synth_frame(&synth, &frame);
		
		unsigned char *rptr = MEM_AUDIO + audio_get_pos();
		
		int i;
		
		
		int nrsamples = synth.pcm.length;
		
		i = 0;
		
		fracptr = 0;
		
		while (nrsamples)
		{
		
			while (wptr == rptr)
			{
				rptr = MEM_AUDIO + audio_get_pos();
			}
			
			while (fracptr >= FRAC)
			{
					// advance source sample
				
				sample[0][0] = sample[1][0];
				sample[0][1] = sample[1][1];
				sample[1][0] = MadFixedToSshort(synth.pcm.samples[0][i]);
				if(MAD_NCHANNELS(&frame.header)==2)
					sample[1][1] = MadFixedToSshort(synth.pcm.samples[1][i]);
				else
					sample[1][1] = sample[1][0];
				++i;
				nrsamples--;
				fracptr -= FRAC;
			}
			
			signed short fs[2];
			
			fs[0] = sample[0][0] + (((sample[1][0] - sample[0][0]) * fracptr) >> FRACSHIFT);
			fs[1] = sample[0][1] + (((sample[1][1] - sample[0][1]) * fracptr) >> FRACSHIFT);
			
			*wptr++ = (fs[0] >> 8);
			*wptr++ = fs[0] & 0xFF;
			
			*wptr++ = (fs[1] >> 8);
			*wptr++ = fs[1] & 0xFF;
			
			fracptr += 7526; // 480 * FRAC / 441;
			if (wptr == end)
				wptr = base;
		}
		dcache_flush(base, (char *)end-(char*)base);
	}
}


/* pad */

void pad_init(void)
{
	*(volatile unsigned long*) 0xCC006400 = 0x00400300; 
	*(volatile unsigned long*) 0xCC00640C = 0x00400300;
	*(volatile unsigned long*) 0xCC006418 = 0x00400300;
	*(volatile unsigned long*) 0xCC006424 = 0x00400300;
	*(volatile unsigned long*) 0xCC006430 = 0x000701f0; 
	*(volatile unsigned long*) 0xCC006438 = 0x80000000; 
	while (!((*(volatile unsigned long*) 0xCC006438)&0x20000000))
	(void)*(volatile unsigned long*)0xCC006434;
	
	*(volatile unsigned long*)0xCC006430 |= 0xF0; // enable all four controller ports
}

void pad_read(int channel, unsigned long *p1, unsigned long *p2)
{
	*p1=*(volatile unsigned long*) 0xCC006404+12*channel;
	*p2=*(volatile unsigned long*) 0xCC006408+12*channel;
}

#define PAD0_A         0x01000000
#define PAD0_B         0x02000000
#define PAD0_X         0x04000000
#define PAD0_Y         0x08000000

#define PAD0_D_LEFT    0x00010000
#define PAD0_D_RIGH    0x00020000
#define PAD0_D_DOWN    0x00040000
#define PAD0_D_UP      0x00080000

int calc_rq(void);

void do_anal(int nr, float val)
{
	int x = nr * 10;
	int y = 400;
	int xs = 8;
	int ys = val * 10.0;
	
	y -= ys;
	
	int ay;
	for (ay = 0; ay < ys; ++ay)
	{
		unsigned long *offset = (unsigned long*)(MEM_FB + (y + ay) * 1280 + x * 4);
		int len = xs;
		while (len--)
			*offset++ = 0xFF80FF80;
	}
}

void video_wait_vsync(void)
{
	while(*(volatile unsigned short*)(0xcc00202C) >= 200);
	while(*(volatile unsigned short*)(0xcc00202C) <  200);
}

void dovisual(void)
{
	float rms[32];
	
	int channel, sample, subband;
	
	for (subband = 0; subband < 32; ++subband)
		rms[subband] = 0;
	
	int samples = MAD_NSBSAMPLES(&frame.header);
			// let's support stereo only...
	if(frame.header.mode!=MAD_MODE_SINGLE_CHANNEL)
		for(channel = 0; channel < 2; channel++)
			for(sample = 0; sample < samples; sample++)
				for(subband=0; subband<32; subband++)
				{
					float s = frame.sbsample[channel][sample][subband] / 16777216.0;
					rms[subband] += s * s;
				}

//	printf("%d samples\n", samples);
	float rmssum = 0;

	for (subband = 0; subband < 32; ++subband)
	{
//		printf("%f\n", rms[subband]);
		rms[subband] = sqrtf(rms[subband]);
		rmssum += rms[subband] / (samples * 2);
		do_anal(subband, log(rms[subband])+7);
	}
}

void clear_screen()
{
	unsigned long *video_fb = (void*)MEM_FB;
	int pixel = 320 * 576;
	while (pixel--)
		*video_fb++ = blit_lookup[0];
}

int mp3_play(int sector, int count, char *filename)
{
	int ret = 1;
	memset(MEM_AUDIO, 0, AUDIO_SIZE);
	
	memset((void*)0x80e00000, 0, 1*1024*1024);

	audio_start(MEM_AUDIO, AUDIO_SIZE);
	current_offset = sector * 2048;
	remaining_size = count;
	
	mad_stream_init(&stream);
	mad_frame_init(&frame);
	mad_synth_init(&synth);
	mad_timer_reset(&timer);
	
	while (remaining_size)
	{
		int p1, p2;
		pad_read(0, &p1, &p2);
		if (p1 & PAD0_B)
		{
			ret = 0;
			break;
		}
		refill_audio_buffer();
		
		video_wait_vsync();
		clear_screen();
		dovisual();
		writex(30, 100, 640-30, font_height, "you're listening to", blit_lookup);
		writex(30, 100 + font_height, 640-30, font_height, filename, blit_lookup);
	}
	audio_stop();
	return ret;
}

void conf_chip(unsigned char c)
{
	exi_select();
	exi_write_word(0xc0000000);
	exi_write_word(c << 24);
	exi_deselect();
}

static inline void draw_border(int size)
{
	int y;
	int x = 100 - 8;
	for (y=100-8; y<100-4; ++y)
		for (x=100-8; x < 100+400+8; x += 2)
			fb[y*320+x/2] = blit_lookup[3];
	int ys = 100+font_height*size;
	
	for (y=ys+4; y<ys+8; ++y)
		for (x=100-8; x < 100+400+8; x += 2)
			fb[y*320+x/2] = blit_lookup[3];

	for (y=100-8; y<ys+8; ++y)
	{
		for (x=100-8; x < 100-4; x += 2)
			fb[y*320+x/2] = blit_lookup[3];
		for (x=400+100+4; x < 400+100+8; x += 2)
			fb[y*320+x/2] = blit_lookup[3];
	}
	
}

int main()
{
	conf_chip(2|1);
	set_video( *(unsigned long*)0x800000CC);
	
	calc_rq();

	console_init(fb, 640, 576, 640*2);

//	pad_init();
	
	init_font();
	
	mtmsr(mfmsr() | 0x8002);
	
	int p = 640 * 576 / 2;
	unsigned long *f = fb;
	while (p--)
		*f++ = blit_lookup[0]; 
	
	write_font(100, 100, "MP3 PLAYER!");
	
//	dvd_do_hack();
	
	set_status("Reading DVD...");

	int sector = 16;
	static unsigned char buffer[2048] __attribute__((aligned(32)));
	
	struct pvd_s *pvd = 0;
	struct pvd_s *svd = 0;
	
	set_status("Reading in CD");
	
	while (sector < 32)
	{
		if (read_sector(buffer, sector))
			fatal_read_error();
		if (!memcmp(((struct pvd_s*)buffer)->id, "\2CD001\1", 8))
		{
			svd = (void*)buffer;
			break;
		}
		++sector;
	}
	
	if (!svd)
	{
		sector = 16;
		while (sector < 32)
		{
			if (read_sector(buffer, sector))
				fatal_read_error();
			
			if (!memcmp(((struct pvd_s*)buffer)->id, "\1CD001\1", 8))
			{
				pvd = (void*)buffer;
				break;
			}
			++sector;
		}
	}
	
	if ((!pvd) && (!svd))
	{
		set_status("no ISO9660 DVD found!");
		while (1);
	}

	files = 0;
	if (svd)
	{
		is_unicode = 1;	
		read_direntry(svd->root_direntry);
	} else
	{
		is_unicode = 0;
		read_direntry(pvd->root_direntry);
	}
	
		// enter root
	read_directory(file[0].sector, file[0].size);

	set_status(is_unicode ? "DVD with Joliet Extension found!" : "DVD with ISO 9660 found!");
	
	int pressed = 0, selection = 0, redraw = 1;
	int offset = 0;

	while (1)
	{
		if (redraw)
		{
			video_wait_vsync();
			clear_screen();
			draw_border(PAGESIZE);
			show_files(offset, selection);
			redraw = 0;
		}
		
		unsigned long p1, p2;
		pad_read(0, &p1, &p2);
		
		int up = (p1 & PAD0_D_UP);
		int down = (p1 & PAD0_D_DOWN);
		int ok = (p1 & PAD0_A);
		
		if (!(up || down || ok))
			pressed = 0;

		if (!pressed)
		{
a:
			if (down)
			{
				selection++;
				
				if (selection == files)
				{
					selection = 0;
					offset = 0;
				}
				
				if ((selection - offset) >= PAGESIZE)
					offset += PAGESIZE;
				redraw = 1;
			} 
			if (up)
			{
				selection--;
				if (selection < 0)
				{
					selection = files - 1;
					offset = selection - PAGESIZE + 1;
				}
				
				if (selection < offset)
					offset -= PAGESIZE;
				
				if (offset < 0)
					offset = 0;
				redraw = 1;
			}
			if (ok)
			{
				if (file[selection].flags & 2) // dir
				{
					read_directory(file[selection].sector, file[selection].size);
					selection = offset = 0;
					redraw = 1;
				} else
				{
					if (mp3_play(file[selection].sector, file[selection].size, file[selection].name))
						if (selection < (files - 1))
						{
							down = 1;
							ok = 1;
							goto a;
						}
					redraw = 1;
				}
			}
		}
		
		if (up || down || ok)
			pressed = 1;
	}

	return 0;
}
